/* content of dxf2html.css */

:root {
    --pp-bg: #20324D; /* background */
    --pp-bg-lt: #3F5577; /* lighter background */
    --pp-tag-color: white; /* general test color */
    --pp-ctrl: orange; /* control tag 0 */
    --pp-group: #A0BFDE; /* group tag 100, 102, ... */
    --pp-link: rgb(250, 195, 150); /* group tag 100, 102, ... */
    --pp-link-hover: orange; /* group tag 100, 102, ... */
}
body {
    font-family: "Open Sans", sans-serif;
    width: 1100px;
    margin-right: auto;
    margin-left: auto;
    color: var(--pp-tag-color);
    background-color: lightgray;
}

.button-bar {
    font-weight: bold;
    padding: 5px;
    color: darkblue;
}

a.link-button {
    background-color: var(--pp-bg-lt);
    color: white;
    font-weight: bold;
    padding: 0 10px 0 10px;
    text-decoration: none;
}

a.link-button:link,
a.link-button:visited,
a.link-button:active,
a.link-button:focus
{
    color: white;
}

a.link-button:hover {
    color: var(--pp-link-hover);
}

/* section, table and entity names as link to the DXF reference
*/

a.dxf-ref-link {
    font-weight: bold;
    text-decoration: none;
}

a.dxf-ref-link:link,
a.dxf-ref-link:visited,
a.dxf-ref-link:active,
a.dxf-ref-link:focus
{
    color: var(--pp-link);
}

a.dxf-ref-link:hover {
    color: var(--pp-link-hover);
}

#section-links {
    margin: 5px;
}

#dxf-filename {
    background-color: var(--pp-bg);
    font-size: xx-large;
    font-weight: bold;
    padding: 5px;
}

#dxf-file {
}

/* DXF sections
*/

.dxf-sections {
}

.dxf-section-name{
    font-size: xx-large;
    font-weight: bold;
    padding: 5px;
}

.dxf-section {
    background-color: var(--pp-bg);
    padding: 5px;
}

/* DXF header
*/

.dxf-header {
    background-color: var(--pp-bg-lt);
    margin: 5px;
    padding: 5px;
}

/* DXF entities
*/
.dxf-entities {
}

.dxf-entity-name{
    font-size: large;
    font-weight: bold;
    padding: 5px;
}

.dxf-entity {
    background-color: slategrey;
    margin: 5px 0 5px 5px;
}

.dxf-blocks,
.dxf-tables {
}

.dxf-block-name,
.dxf-table-name {
    font-size: x-large;
    font-weight: bold;
    padding: 5px;
}

.dxf-block,
.dxf-table{
    background-color: #4c8b8d;
    margin: 5px 5px 0 5px;
    padding: 5px 0 0 5px;
}

/* DXF tags and header vars
*/
.dxf-tags {
    background-color: var(--pp-bg-lt);
    padding: 5px;
    color: var(--pp-tag-color)
}

.dxf-tag,
.hdr-var {
    font-family: "Source Code Pro", monospace;
    font-size: medium;
}

.tag-code,
.var-type {
}

.tag-value {
}

/* DXF handles and pointers as links
*/
.tag-link {
    font-weight: bold;
    text-decoration: none;
}
.tag-link:link,
.tag-link:visited,
.tag-link:active,
.tag-link:focus
{
    color: var(--pp-link);
}
.tag-link:hover {
    color: var(--pp-link-hover);
}

.tag-group-marker {
    color: var(--pp-group);
}
